"""Results and postprocessing for the :mod:`hepi` package."""
from .util import DictData
import numpy as np
from uncertainties import unumpy
from smpl import plot
import lhapdf
import warnings

required_numerical_uncertainty_factor = 10
"""If the numerical uncertainty is :attr:`required_numerical_uncertainty_factor` times higher than the scale or pdf uncertainty a warning is shown."""


class Result(DictData):
    """
    General result class.

    Attributes:
        LO (:obj:`double`): Leading Order result. Defaults to None.
        NLO (:obj:`double`): Next-to-Leading Order result. Defaults to None.
        NLO_PLUS_NLL (:obj:`double`): Next-to-Leading Order plus Next-to-Leading Logarithm result. Defaults to None.
        K_LO (:obj:`double`): LO divided by LO.
        K_NLO (:obj:`double`): NLO divided by LO result.
        K_NLO_PLUS_NLL (:obj:`double`): NLO+NLL divided by LO.
        K_aNNLO_PLUS_NNLL (:obj:`double`): aNNLO+NNLL divided by LO.
        NLO_PLUS_NLL_OVER_NLO (:obj:`double`): NLO+NLL divided by NLO.
        aNNLO_PLUS_NNLL_OVER_NLO (:obj:`double`): aNNLO+NNLL divided by NLO.
    """

    def __init__(self,
                 lo=None,
                 nlo=None,
                 nlo_plus_nll=None,
                 annlo_plus_nnll=None):
        """
        Sets given and computes dependent ``Attributes``.

        Args:
            lo (:obj:`double`): corresponds to :attr:`LO`.
            nlo (:obj:`double`): corresponds to :attr:`NLO`.
            nlo_plus_nll (:obj:`double`): corresponds to :attr:`NLO_PLUS_NLL`.
            annlo_plus_nnll (:obj:`double`): corresponds to :attr:`aNNLO_PLUS_NNLL`.
        """
        self.LO = lo
        self.NLO = nlo
        self.NLO_PLUS_NLL = nlo_plus_nll
        self.aNNLO_PLUS_NNLL = annlo_plus_nnll
        if lo is not None and lo != 0:
            self.K_LO = lo / lo
        else:
            self.K_LO = None
        if nlo is not None and lo != 0:
            self.K_NLO = nlo / lo
        else:
            self.K_NLO = None
        if nlo_plus_nll is not None and lo != 0:
            self.K_NLO_PLUS_NLL = nlo_plus_nll / lo
        else:
            self.K_NLO_PLUS_NLL = None

        if nlo_plus_nll is not None and nlo != 0:
            self.NLO_PLUS_NLL_OVER_NLO = nlo_plus_nll / nlo
        else:
            self.NLO_PLUS_NLL_OVER_NLO = None

        if annlo_plus_nnll is not None and lo != 0:
            self.K_aNNLO_PLUS_NNLL = annlo_plus_nnll / lo
        else:
            self.K_aNNLO_PLUS_NNLL = None

        if annlo_plus_nnll is not None and nlo != 0:
            self.aNNLO_PLUS_NNLL_OVER_NLO = annlo_plus_nnll / nlo
        else:
            self.aNNLO_PLUS_NNLL_OVER_NLO = None


# TODO detect which errors also for scales
def pdf_errors(li,
               dl,
               ordernames=["LO", "NLO", "aNNLO_PLUS_NNLL"],
               confidence_level=90):
    """
    Just like `pdf_error` but over a list of ordernames.
    """
    r_dl = dl
    for o in ordernames:
        r_dl = pdf_error(li, r_dl, o, confidence_level=confidence_level)

    return r_dl


def pdf_error(li, dl, ordername="LO", confidence_level=90):
    """
    Computes Parton Density Function (PDF) uncertainties through :func:`lhapdf.set.uncertainty`.

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.
        ordername (`str`): Name of the order.
        confidence_level (:obj:`double`): Confidence Level for PDF uncertainty

    Returns:
        :obj:`dict`: Modified `dl` with new `LO`/`NLO`/`NLO_PLUS_NLL` _ `PDF`/`PDF_CENTRAL`/`PDF_ERRPLUS`/`PDF_ERRMINUS`/`PDF_ERRSYM` entries.
            - `LO`/`NLO`/`NLO_PLUS_NLL` _ `PDF` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [
        attr for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]

    dl[ordername + "_PDF"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_CENTRAL"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRSYM"] = np.array([None] * len(dl["pdfset_nlo"]))

    for i in range(len(dl["pdfset_nlo"])):
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][
                i] == 1.0:
            pdfset = lhapdf.getPDFSet(dl["pdf_nlo"][i])
            pdfs = [0.0] * pdfset.size
            for j in range(len(dl["pdfset_nlo"])):
                same = True
                for s in members:
                    if not (
                            dl[s][i] == dl[s][j]
                    ) and s != "pdfset_nlo" and s != "precision" and s != "max_iters":
                        same = False
                if same:
                    pdfs[dl["pdfset_nlo"][j]] = j

        # lo_unc = pdfset.uncertainty(
        #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
            if ordername == "LO":
                dl.loc[i, "LO_PDF_CENTRAL"] = plot.unv(dl["LO"][i])
                dl.loc[i, "LO_PDF_ERRPLUS"] = 0.0
                dl.loc[i, "LO_PDF_ERRMINUS"] = 0.0
                dl.loc[i, "LO_PDF_ERRSYM"] = 0.0
            else:
                nlo_unc = pdfset.uncertainty(
                    [plot.unv(dl["NLO"][k]) for k in pdfs], confidence_level)
                dl.loc[i, ordername + "_PDF_CENTRAL"] = nlo_unc.central
                dl.loc[i, ordername + "_PDF_ERRPLUS"] = nlo_unc.errplus
                dl.loc[i, ordername + "_PDF_ERRMINUS"] = -nlo_unc.errminus
                dl.loc[i, ordername + "_PDF_ERRSYM"] = nlo_unc.errsymm
            #TODO error sym to minus and plus
            if ordername != "LO":
                if (plot.usd(dl[ordername][i]) *
                        required_numerical_uncertainty_factor >
                        dl[ordername + "_PDF_ERRPLUS"][i]
                        or plot.usd(dl[ordername][i]) *
                        required_numerical_uncertainty_factor >
                        -dl[ordername + "_PDF_ERRMINUS"][i]):
                    rel = plot.unv(dl[ordername][i])
                    warnings.warn(
                        "too bad numerical precision vs pdf @ " + ordername +
                        " num: " +
                        str(plot.usd(dl[ordername][i]) / rel * 100.) +
                        "% vs " +
                        str(dl[ordername + "_PDF_ERRPLUS"][i] / rel * 100.) +
                        "% to pdf: " +
                        str(dl[ordername + "_PDF_ERRMINUS"][i] / rel * 100.) +
                        "%", RuntimeWarning)

    mask = dl[ordername + "_PDF_CENTRAL"].notnull()
    dl.loc[mask, ordername + "_PDF"] = unumpy.uarray(
        plot.unv(dl[ordername + "_PDF_CENTRAL"][mask]) +
        dl[ordername + "_PDF_ERRPLUS"][mask] / 2. +
        dl[ordername + "_PDF_ERRMINUS"][mask] / 2.,
        (+dl[ordername + "_PDF_ERRPLUS"][mask] -
         dl[ordername + "_PDF_ERRMINUS"][mask]) / 2.)

    return dl


def pdf_error_old(li, dl, confidence_level=90):
    """
    Computes Parton Density Function (PDF) uncertainties through :func:`lhapdf.set.uncertainty`.

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.
        confidence_level (:obj:`double`): Confidence Level for PDF uncertainty

    Returns:
        :obj:`dict`: Modified `dl` with new `LO`/`NLO`/`NLO_PLUS_NLL` _ `PDF`/`PDF_CENTRAL`/`PDF_ERRPLUS`/`PDF_ERRMINUS`/`PDF_ERRSYM` entries.
            - `LO`/`NLO`/`NLO_PLUS_NLL` _ `PDF` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [
        attr for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]
    dl["LO_PDF"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_PDF_CENTRAL"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_PDF_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_PDF_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_PDF_ERRSYM"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PDF"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PDF_CENTRAL"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PDF_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PDF_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PDF_ERRSYM"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_CENTRAL"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_ERRSYM"] = np.array([None] * len(dl["pdfset_nlo"]))
    for i in range(len(dl["pdfset_nlo"])):
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][
                i] == 1.0:
            pdfset = lhapdf.getPDFSet(dl["pdf_nlo"][i])
            pdfs = [0.0] * pdfset.size
            for j in range(len(dl["pdfset_nlo"])):
                same = True
                for s in members:
                    if not (
                            dl[s][i] == dl[s][j]
                    ) and s != "pdfset_nlo" and s != "precision" and s != "max_iters":
                        same = False
                if same:
                    pdfs[dl["pdfset_nlo"][j]] = j

            dl["LO_PDF_CENTRAL"][i] = plot.unv(dl["LO"][i])
            dl["LO_PDF_ERRPLUS"][i] = 0.0
            dl["LO_PDF_ERRMINUS"][i] = 0.0
            dl["LO_PDF_ERRSYM"][i] = 0.0
            # lo_unc = pdfset.uncertainty(
            #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
            nlo_unc = pdfset.uncertainty(
                [plot.unv(dl["NLO"][k]) for k in pdfs], confidence_level)
            dl["NLO_PDF_CENTRAL"][i] = nlo_unc.central
            dl["NLO_PDF_ERRPLUS"][i] = nlo_unc.errplus
            dl["NLO_PDF_ERRMINUS"][i] = -nlo_unc.errminus
            dl["NLO_PDF_ERRSYM"][i] = nlo_unc.errsymm
            #TODO error sym to minus and plus
            if (plot.usd(dl["NLO"][i]) * required_numerical_uncertainty_factor
                    > dl["NLO_PDF_ERRPLUS"][i] or plot.usd(dl["NLO"][i]) *
                    required_numerical_uncertainty_factor >
                    -dl["NLO_PDF_ERRMINUS"][i]):
                warnings.warn("too low numerical nlo precision vs pdf.",
                              RuntimeWarning)

            nlo_plus_nll_unc = pdfset.uncertainty(
                [plot.unv(dl["NLO_PLUS_NLL"][k]) for k in pdfs], 90)
            dl["NLO_PLUS_NLL_PDF_CENTRAL"][i] = nlo_plus_nll_unc.central
            dl["NLO_PLUS_NLL_PDF_ERRPLUS"][i] = nlo_plus_nll_unc.errplus
            dl["NLO_PLUS_NLL_PDF_ERRMINUS"][i] = -nlo_plus_nll_unc.errminus
            dl["NLO_PLUS_NLL_PDF_ERRSYM"][i] = nlo_plus_nll_unc.errsymm
            #TODO error sym to minus and plus
            if (plot.usd(dl["NLO_PLUS_NLL"][i]) *
                    required_numerical_uncertainty_factor >
                    dl["NLO_PLUS_NLL_PDF_ERRPLUS"][i]
                    or plot.usd(dl["NLO_PLUS_NLL"][i]) *
                    required_numerical_uncertainty_factor >
                    -dl["NLO_PLUS_NLL_PDF_ERRMINUS"][i]):
                warnings.warn(
                    "too low numerical nlo_plus_nll precision vs pdf.",
                    RuntimeWarning)

    mask = dl["LO_PDF_CENTRAL"] != np.array(None)
    dl["LO_PDF"][mask] = unumpy.uarray(
        plot.unv(dl["LO_PDF_CENTRAL"][mask]) +
        dl["LO_PDF_ERRPLUS"][mask] / 2. + dl["LO_PDF_ERRMINUS"][mask] / 2.,
        (+dl["LO_PDF_ERRPLUS"][mask] - dl["LO_PDF_ERRMINUS"][mask]) / 2.)

    mask = dl["NLO_PDF_CENTRAL"] != np.array(None)
    dl["NLO_PDF"][mask] = unumpy.uarray(
        plot.unv(dl["NLO_PDF_CENTRAL"][mask]) +
        dl["NLO_PDF_ERRPLUS"][mask] / 2. + dl["NLO_PDF_ERRMINUS"][mask] / 2.,
        (+dl["NLO_PDF_ERRPLUS"][mask] - dl["NLO_PDF_ERRMINUS"][mask]) / 2.)

    mask = dl["NLO_PLUS_NLL_PDF_CENTRAL"] != np.array(None)
    dl["NLO_PLUS_NLL_PDF"][mask] = unumpy.uarray(
        plot.unv(dl["NLO_PLUS_NLL_PDF_CENTRAL"][mask]) +
        dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask] / 2. +
        dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask] / 2.,
        (+dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask] -
         dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask]) / 2.)
    return dl


def scale_errors(li, dl, ordernames=["LO", "NLO", "aNNLO_PLUS_NNLL"]):
    """
    Just like `scale_error` but over a list of ordernames.
    """
    r_dl = dl
    for o in ordernames:
        r_dl = scale_error(li, r_dl, o)

    return r_dl


def scale_error(li, dl, ordername="LO"):
    """
    Computes seven-point scale uncertainties from the results where the renormalization and factorization scales are varied by factors of 2 and  relative factors of four are excluded (cf. :meth:`seven_point_scan`).

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `LO`/`NLO`/`NLO_PLUS_NLL` _ `SCALE`/`SCALE_ERRPLUS`/`SCALE_ERRMINUS`/`SCALE_ERRSYM` entries.
            - `LO`/`NLO`/`NLO_PLUS_NLL` _ `SCALE` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [
        attr for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]
    dl[ordername + "_SCALE"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_SCALE_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_SCALE_ERRMINUS"] = np.array([None] *
                                                 len(dl["pdfset_nlo"]))

    for i in range(len(dl["pdfset_nlo"])):
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][
                i] == 1.0:
            scales = []
            for j in range(len(dl["pdfset_nlo"])):
                same = True
                for s in members:
                    if not (
                            dl[s][i] == dl[s][j]
                    ) and s != "mu_f" and s != "mu_r" and s != "precision" and s != "max_iters":
                        same = False
                if same:
                    scales.append(j)

            # lo_unc = pdfset.uncertainty(
            #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
            dl.loc[i, ordername + "_SCALE_ERRPLUS"] = np.max(
                [plot.unv(dl[ordername][k])
                 for k in scales]) - plot.unv(dl[ordername][i])
            dl.loc[i, ordername + "_SCALE_ERRMINUS"] = np.min(
                [plot.unv(dl[ordername][k])
                 for k in scales]) - plot.unv(dl[ordername][i])
            if (plot.usd(dl[ordername][i]) *
                    required_numerical_uncertainty_factor >
                    dl[ordername + "_SCALE_ERRPLUS"][i]
                    or plot.usd(dl[ordername][i]) *
                    required_numerical_uncertainty_factor >
                    -dl[ordername + "_SCALE_ERRMINUS"][i]):
                rel = plot.unv(dl[ordername][i])
                warnings.warn(
                    "too bad numerical precision vs scale @ num:" + ordername +
                    " " + str(plot.usd(dl[ordername][i]) / rel * 100.) +
                    "% vs scale:" +
                    str(dl[ordername + "_SCALE_ERRPLUS"][i] / rel * 100.) +
                    "% to " +
                    str(dl[ordername + "_SCALE_ERRMINUS"][i] / rel * 100.) +
                    "%", RuntimeWarning)

    mask = dl[ordername + "_SCALE_ERRPLUS"].notnull()
    dl.loc[mask, ordername + "_SCALE"] = unumpy.uarray(
        plot.unv(dl[ordername][mask]) +
        dl[ordername + "_SCALE_ERRPLUS"][mask] / 2. +
        dl[ordername + "_SCALE_ERRMINUS"][mask] / 2.,
        (+dl[ordername + "_SCALE_ERRPLUS"][mask] -
         dl[ordername + "_SCALE_ERRMINUS"][mask]) / 2.)

    return dl


def scale_error_old(li, dl):
    """
    Computes seven-point scale uncertainties from the results where the renormalization and factorization scales are varied by factors of 2 and  relative factors of four are excluded (cf. :meth:`seven_point_scan`).

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `LO`/`NLO`/`NLO_PLUS_NLL` _ `SCALE`/`SCALE_ERRPLUS`/`SCALE_ERRMINUS`/`SCALE_ERRSYM` entries.
            - `LO`/`NLO`/`NLO_PLUS_NLL` _ `SCALE` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [
        attr for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]
    dl["LO_SCALE"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_SCALE_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_SCALE_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_SCALE"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_SCALE_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_SCALE_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_SCALE"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_SCALE_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_SCALE_ERRMINUS"] = np.array([None] *
                                                 len(dl["pdfset_nlo"]))
    for i in range(len(dl["pdfset_nlo"])):
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][
                i] == 1.0:
            scales = []
            for j in range(len(dl["pdfset_nlo"])):
                same = True
                for s in members:
                    if not (
                            dl[s][i] == dl[s][j]
                    ) and s != "mu_f" and s != "mu_r" and s != "precision" and s != "max_iters":
                        same = False
                if same:
                    scales.append(j)

            # lo_unc = pdfset.uncertainty(
            #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
            dl["LO_SCALE_ERRPLUS"][i] = np.max(
                [plot.unv(dl["LO"][k])
                 for k in scales]) - plot.unv(dl["LO"][i])
            dl["LO_SCALE_ERRMINUS"][i] = np.min(
                [plot.unv(dl["LO"][k])
                 for k in scales]) - plot.unv(dl["LO"][i])
            if (plot.usd(dl["LO"][i]) * required_numerical_uncertainty_factor >
                    dl["LO_SCALE_ERRPLUS"][i] or plot.usd(dl["LO"][i]) *
                    required_numerical_uncertainty_factor >
                    -dl["LO_SCALE_ERRMINUS"][i]):
                warnings.warn("too low numerical lo precision vs scale.",
                              RuntimeWarning)

            dl["NLO_SCALE_ERRPLUS"][i] = np.max(
                [plot.unv(dl["NLO"][k])
                 for k in scales]) - plot.unv(dl["NLO"][i])
            dl["NLO_SCALE_ERRMINUS"][i] = np.min(
                [plot.unv(dl["NLO"][k])
                 for k in scales]) - plot.unv(dl["NLO"][i])
            if (plot.usd(dl["NLO"][i]) * required_numerical_uncertainty_factor
                    > dl["NLO_SCALE_ERRPLUS"][i] or plot.usd(dl["NLO"][i]) *
                    required_numerical_uncertainty_factor >
                    -dl["NLO_SCALE_ERRMINUS"][i]):
                warnings.warn("too low numerical nlo precision vs scale.",
                              RuntimeWarning)

            dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][i] = np.max(
                [plot.unv(dl["NLO_PLUS_NLL"][k])
                 for k in scales]) - plot.unv(dl["NLO_PLUS_NLL"][i])
            dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][i] = np.min(
                [plot.unv(dl["NLO_PLUS_NLL"][k])
                 for k in scales]) - plot.unv(dl["NLO_PLUS_NLL"][i])
            if (plot.usd(dl["NLO_PLUS_NLL"][i]) *
                    required_numerical_uncertainty_factor >
                    dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][i]
                    or plot.usd(dl["NLO_PLUS_NLL"][i]) *
                    required_numerical_uncertainty_factor >
                    -dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][i]):
                warnings.warn(
                    "too low numerical nlo_plus_nll precision vs scale.",
                    RuntimeWarning)

    mask = dl["LO_SCALE_ERRPLUS"] != np.array(None)
    dl["LO_SCALE"][mask] = unumpy.uarray(
        plot.unv(dl["LO"][mask]) + dl["LO_SCALE_ERRPLUS"][mask] / 2. +
        dl["LO_SCALE_ERRMINUS"][mask] / 2.,
        (+dl["LO_SCALE_ERRPLUS"][mask] - dl["LO_SCALE_ERRMINUS"][mask]) / 2.)

    mask = dl["NLO_SCALE_ERRPLUS"] != np.array(None)
    dl["NLO_SCALE"][mask] = unumpy.uarray(
        plot.unv(dl["NLO"][mask]) + dl["NLO_SCALE_ERRPLUS"][mask] / 2. +
        dl["NLO_SCALE_ERRMINUS"][mask] / 2.,
        (+dl["NLO_SCALE_ERRPLUS"][mask] - dl["NLO_SCALE_ERRMINUS"][mask]) / 2.)

    mask = dl["NLO_PLUS_NLL_SCALE_ERRPLUS"] != np.array(None)
    dl["NLO_PLUS_NLL_SCALE"][mask] = unumpy.uarray(
        plot.unv(dl["NLO_PLUS_NLL"][mask]) +
        dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask] / 2. +
        dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask] / 2.,
        (+dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask] -
         dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask]) / 2.)

    return dl


def combine_errors(dl, ordernames=["LO", "NLO", "aNNLO_PLUS_NNLL"]):
    """
    Just like `combine_error` but over a list of ordernames.
    """
    r_dl = dl
    for o in ordernames:
        r_dl = combine_error(r_dl, o)

    return r_dl


def combine_error(dl: dict, ordername="LO"):
    """
    Combines seven-point scale uncertainties and pdf uncertainties from the results by Pythagorean addition.

    Note:
        Running :func:`scale_errors` and :func:`pdf_errors` before is necessary.

    Args:
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `LO`/`NLO`/`NLO_PLUS_NLL` _ `COMBINED`/`ERRPLUS`/`ERRMINUS` entries.
            - `LO`/`NLO`/`NLO_PLUS_NLL` _ `COMBINED` contains a symmetrized :mod:`uncertainties` object.
    """
    dl[ordername + "_NOERR"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_COMBINED"] = np.array([None] * len(dl["pdfset_nlo"]))

    mask = dl[ordername + "_PDF_CENTRAL"].notnull()
    dl.loc[mask, ordername + "_NOERR"] = plot.unv(dl[ordername +
                                                     ""][mask]).astype(float)
    dl.loc[mask, ordername + "_ERRPLUS"] = np.sqrt(
        dl[ordername + "_PDF_ERRPLUS"][mask].astype(float)**2 +
        dl[ordername + "_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl.loc[mask, ordername + "_ERRMINUS"] = -np.sqrt(
        dl[ordername + "_PDF_ERRMINUS"][mask].astype(float)**2 +
        dl[ordername + "_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl.loc[mask, ordername + "_COMBINED"] = unumpy.uarray(
        plot.unv(dl[ordername + ""][mask]) +
        dl[ordername + "_ERRPLUS"][mask] / 2. +
        dl[ordername + "_ERRMINUS"][mask] / 2.,
        +dl[ordername + "_ERRPLUS"][mask] - dl[ordername + "_ERRMINUS"][mask])

    return dl


def combine_errors_old(dl: dict):
    """
    Combines seven-point scale uncertainties and pdf uncertainties from the results by Pythagorean addition.

    Note:
        Running :func:`scale_errors` and :func:`pdf_errors` before is necessary.

    Args:
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `LO`/`NLO`/`NLO_PLUS_NLL` _ `COMBINED`/`ERRPLUS`/`ERRMINUS` entries.
            - `LO`/`NLO`/`NLO_PLUS_NLL` _ `COMBINED` contains a symmetrized :mod:`uncertainties` object.
    """
    dl["LO_NOERR"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["LO_COMBINED"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_NOERR"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_COMBINED"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_NOERR"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_COMBINED"] = np.array([None] * len(dl["pdfset_nlo"]))

    mask = dl["LO_PDF_CENTRAL"] != np.array(None)
    dl["LO_NOERR"][mask] = plot.unv(dl["LO"][mask]).astype(float)
    dl["LO_ERRPLUS"][mask] = np.sqrt(
        dl["LO_PDF_ERRPLUS"][mask].astype(float)**2 +
        dl["LO_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl["LO_ERRMINUS"][mask] = -np.sqrt(dl["LO_PDF_ERRMINUS"][mask].astype(
        float)**2 + dl["LO_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl["LO_COMBINED"][mask] = unumpy.uarray(
        plot.unv(dl["LO"][mask]) + dl["LO_ERRPLUS"][mask] / 2. +
        dl["LO_ERRMINUS"][mask] / 2.,
        +dl["LO_ERRPLUS"][mask] - dl["LO_ERRMINUS"][mask])

    mask = dl["NLO_PDF_CENTRAL"] != np.array(None)
    dl["NLO_NOERR"][mask] = plot.unv(dl["NLO"][mask]).astype(float)
    dl["NLO_ERRPLUS"][mask] = np.sqrt(
        dl["NLO_PDF_ERRPLUS"][mask].astype(float)**2 +
        dl["NLO_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl["NLO_ERRMINUS"][mask] = -np.sqrt(dl["NLO_PDF_ERRMINUS"][mask].astype(
        float)**2 + dl["NLO_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl["NLO_COMBINED"][mask] = unumpy.uarray(
        plot.unv(dl["NLO"][mask]) + dl["NLO_ERRPLUS"][mask] / 2. +
        dl["NLO_ERRMINUS"][mask] / 2.,
        +dl["NLO_ERRPLUS"][mask] - dl["NLO_ERRMINUS"][mask])

    mask = dl["NLO_PLUS_NLL_PDF_CENTRAL"] != np.array(None)
    dl["NLO_PLUS_NLL_NOERR"][mask] = plot.unv(
        dl["NLO_PLUS_NLL"][mask]).astype(float)
    dl["NLO_PLUS_NLL_ERRPLUS"][mask] = np.sqrt(
        dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask].astype(float)**2 +
        dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl["NLO_PLUS_NLL_ERRMINUS"][mask] = -np.sqrt(
        dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask].astype(float)**2 +
        dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl["NLO_PLUS_NLL_COMBINED"][mask] = unumpy.uarray(
        plot.unv(dl["NLO_PLUS_NLL"][mask]) +
        dl["NLO_PLUS_NLL_ERRPLUS"][mask] / 2. +
        dl["NLO_PLUS_NLL_ERRMINUS"][mask] / 2.,
        +dl["NLO_PLUS_NLL_ERRPLUS"][mask] - dl["NLO_PLUS_NLL_ERRMINUS"][mask])

    return dl



from logging import warning
import os
import subprocess
from subprocess import Popen, PIPE
from typing import List
import warnings
from hepi.input import Input, Order, get_input_dir, get_output_dir, get_pre
from hepi.results import Result
from hepi.util import DL2DF, LD2DL, DictData, namehash
from smpl.parallel import par
import time
import tqdm


class RunParam(DictData):
    """Abstract class that is similar to a dictionary but with fixed keys."""

    def __init__(self,
                 skip: bool = False,
                 in_file: str = None,
                 out_file: str = None,
                 execute: str = None,
                 name: str = None):
        self.name = name
        self.skip = skip
        self.in_file = in_file
        self.out_file = out_file
        self.execute = execute


class Runner:

    def __init__(self,
                 path: str,
                 in_dir: str = None,
                 out_dir: str = None,
                 pre=None):
        self.path = path
        if in_dir is None:
            self.in_dir = get_input_dir()
        else:
            self.in_dir = in_dir
        if out_dir is None:
            self.out_dir = get_output_dir()
        else:
            self.out_dir = out_dir
        if pre is None:
            self.pre = get_pre()
        else:
            self.pre = pre

    def orders(self) -> List[Order]:
        """List of supported Orders in this runner."""
        return [e.value for e in Order]

    def get_name(self) -> str:
        """Returns the name of the runner."""
        return type(self).__name__

    def get_version(self) -> str:
        return "?"

    def _sub_run(self, coms: List[str]) -> str:
        process = Popen(coms, stdout=PIPE)
        (output, err) = process.communicate()
        exit_code = process.wait()
        if exit_code != 0:
            return err.decode()
        else:
            return output.decode()

    def _check_path(self) -> bool:
        """Checks if the passed path is valid."""
        return True

    def _prepare(self, p: Input, **kwargs) -> RunParam:
        skip_ = kwargs["skip"]
        d = p.__dict__
        d["runner"] = str(type(self).__name__) + "-" + self.get_version(
        )  # TODO re add version, but removed for reusable hashing!
        name = namehash("_".join("".join(str(_[0]) + "_" + str(_[1]))
                                 for _ in d.items()).replace("/", "-"))
        #print(name)
        skip = False
        if skip_ and os.path.isfile(self.get_output_dir() + name +
                                    ".out") and self._is_valid(
                                        self.get_output_dir() + name + ".out",
                                        p, d, **kwargs):
            print(".", end='')
            skip = True
        else:
            print('|', end='')
        return RunParam(execute=self.get_output_dir() + name + ".sh",
                        in_file=self.get_output_dir() + name + ".in",
                        out_file=self.get_output_dir() + name + ".out",
                        skip=skip,
                        name=name)

    def _check_input(self, param: Input, **kwargs) -> bool:
        if param.order not in self.orders():
            warnings.warn("Order " + str(param.order) + " not supported in " +
                          type(self).__name__)
            return False
        return True

    def _prepare_all(self,
                     params: List[Input],
                     skip=True,
                     **kwargs) -> List[RunParam]:
        ret = []
        for p in params:
            if not self._check_input(p):
                warnings.warn("Check input failed.")
                return []
            ret.append(self._prepare(p, skip=skip, **kwargs))
        return ret

    def run(self,
            params: List[Input],
            skip=True,
            parse=True,
            parallel=True,
            sleep=0,
            run=True,
            ignore_error=False,
            **kwargs):
        """
		Run the passed list of parameters.

		Args:
		    params (:obj:`list` of :class:`hepi.Input`): All parameters that should be executed/queued.
		    skip (bool): True means stored runs will be skipped. Else the are overwritten.
		    parse (bool): Parse the results. 
		        This is not the prefered cluster/parallel mode, as there the function only queues the job.
		    parallel (bool): Run jobs in parallel.
		    sleep (int): Sleep seconds after starting job.
            run (bool): Actually start/queue runner.
            ignore_error (bool): Continue instead of raising Exceptions. Also ignores hash collisions.

		Returns:
		    :obj:`pd.DataFrame` : combined dataframe of results and parameters. The dataframe is empty if `parse` is set to False.

		"""
        if not self._check_path():
            warnings.warn("The path is not valid for " + self.get_name())
            if not ignore_error:
                raise RuntimeError("The path is not valid for " +
                                   self.get_name())
        rps = self._prepare_all(params,
                                parse=parse,
                                skip=skip,
                                ignore_error=ignore_error,
                                **kwargs)
        print("= " + str(len(params)) + " jobs")
        if sleep is None:
            sleep = 0 if parse else 5
        if run:
            self._run(rps,
                      wait=parse,
                      parallel=parallel,
                      sleep=sleep,
                      **kwargs)
        if parse:
            outs = LD2DL(rps)["out_file"]
            results = self.parse(outs)
            rdl = LD2DL(results)
            pdl = LD2DL(params)
            return DL2DF({**rdl, **pdl})
        return DL2DF({})

    def _run(self,
             rps: List[RunParam],
             wait=True,
             parallel=True,
             sleep=0,
             **kwargs):
        """
		Runs Runner per :class:`RunParams`.
	
		Args:
		    rps (:obj:`list` of :class:`RunParams`): Extended run parameters.
		    bar (bool): Enable info bar.
		    wait (bool): Wait for parallel runs to finish.
		    sleep (int): Sleep seconds after starting subprocess.
		    parallel (bool): Run jobs in parallel.
	
		Returns:
		    :obj:`list` of int: return codes from jobs if `no_parse` is False.
		"""
        # get cluster or niceness prefix
        template = self.get_pre() + " " + "{}"

        # Run commands in parallel
        processes = []

        for rp in rps:
            if not rp.skip:
                command = template.format(rp.execute)
                process = subprocess.Popen(command, shell=True)
                processes.append(process)
                if not parallel:
                    process.wait()
                # Forced delay to prevent overloading clusters when registering jobs
                time.sleep(sleep)

        if wait:
            # Collect statuses
            output = [p.wait() for p in processes]
            return output
        return []

    def _is_valid(self, file: str, p: Input, d, **kwargs) -> bool:
        """
		Verifies that a file is a complete output.
	
		Args:
		    file (str): File path to be parsed.
		    p (:class:`hepi.Input`): Onput parameters.
		    d (:obj:`dict`): Param dictionary.
	
		Returns:
		    bool : True if `file` could be parsed.
		"""
        res = self._parse_file(file)
        if res.LO is None and p.order is Order.LO:
            return False
        if res.NLO is None and p.order is Order.NLO:
            return False
        if res.NLO_PLUS_NLL is None and p.order is Order.NLO_PLUS_NLL:
            return False
        if res.aNNLO_PLUS_NNLL is None and p.order is Order.aNNLO_PLUS_NNLL:
            return False
        return True

    def parse(self, outputs: List[str]) -> List[Result]:
        """
		Parses Resummino output files and returns List of Results.
	
		Args:
		    outputs (:obj:`list` of `str`): List of the filenames to be parsed.
	
		Returns:
		    :obj:`list` of :class:`hepi.resummino.result.ResumminoResult`
	
		"""
        rsl = []
        #for r in par(self._parse_file, outputs):
        #    rsl.append(r)

        # parallelized opens to many files at times
        for o in tqdm.tqdm(outputs):
            rsl.append(self._parse_file(o))
        return rsl

    def _parse_file(self, file: str) -> Result:
        """
		Extracts results from an output file.

		Args:
		    file (str): File path to be parsed.

		Returns:
		    :class:`Result` : If a value is not found in the file None is used.

		"""
        return None

    def get_path(self) -> str:
        """
		Get the Runner path.

		Returns:
		    str: current Runner path.
		"""
        return self.path

    def get_input_dir(self) -> str:
        """
		Get the input directory.

		Returns:
		    str: :attr:`in_dir`
		"""
        return self.in_dir

    def get_output_dir(self) -> str:
        """
		Get the input directory.

		Returns:
		    str: :attr:`out_dir`
		"""
        return self.out_dir

    def get_pre(self) -> str:
        """
		Gets the command prefix. 

		Returns:
		    str: :attr:`pre`
		"""
        return self.pre

    def set_path(self, p: str):
        """
		Set the path to the Runner folder containing the binary in './bin' or './build/bin'.

		Args:
		    p (str): new path.
		"""
        if os.path.isdir(p):
            self.path = p + ("/" if p[-1] != "/" else "")
        self.path = p

    def set_input_dir(self, indir: str):
        """
		Sets the input directory.

		Args:
		    indir (str): new input directory.
		"""
        self.in_dir = indir

    def set_output_dir(self, outdir: str, create: bool = True):
        """
		Sets the output directory.

		Args:
		    outdir (str): new output directory.
			create (bool): create directory if not existing.
		"""
        if create:
            os.makedirs(outdir, exist_ok=True)
        self.out_dir = outdir

    def set_pre(self, ppre: str):
        """
		Sets the command prefix. 

		Args:
		    ppre (str): new command prefix.
		"""
        self.pre = ppre
